import platform
import socket
import subprocess
import sys
import uuid


def execute(cmd):
    try:
        return subprocess.run(
            cmd, shell=True, capture_output=True, check=True, encoding="utf-8"
        ).stdout.strip()
    except:
        return None


def read_win_registry(registry, key):
    try:
        from winregistry import WinRegistry

        with WinRegistry() as reg:
            return reg.read_entry(registry, key).value.strip()
    except:
        return None


def read_file(path):
    try:
        with open(path) as f:
            return f.read()
    except:
        return None


class HardwareIdProvider:
    def get_id(self):
        return str(uuid.getnode())

    def get_os_ver(self):
        return platform.platform()

    def get_hostname(self):
        return platform.node()

    def get_ip(self):
        return socket.gethostbyname(self.get_hostname())

    def get_is_vm(self):
        return False

    def get_vm_info(self):
        return None

    def get_mac_address(self):
        return ":".join(("%012X" % uuid.getnode())[i : i + 2] for i in range(0, 12, 2))

    def get_request_id(self):
        return str(uuid.uuid4())


class PlatformIdProvider(HardwareIdProvider):
    def get_id(self):
        id = None

        if sys.platform == "darwin":
            id = execute(
                "ioreg -d2 -c IOPlatformExpertDevice | awk -F\\\" '/IOPlatformUUID/{print $(NF-1)}'"
            )

        if (
            sys.platform == "win32"
            or sys.platform == "cygwin"
            or sys.platform == "msys"
        ):
            id = read_win_registry(
                "HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Cryptography", "MachineGuid"
            )
            if not id:
                id = execute("wmic csproduct get uuid").split("\n")[2].strip()

        if sys.platform.startswith("linux"):
            id = read_file("/var/lib/dbus/machine-id")
            if not id:
                id = read_file("/etc/machine-id")

        if sys.platform.startswith("openbsd") or sys.platform.startswith("freebsd"):
            id = read_file("/etc/hostid")
            if not id:
                id = execute("kenv -q smbios.system.uuid")

        if not id:
            id = super().get_id()

        return id



from requests import HTTPError


class ClientError(HTTPError):
    def __init__(
        self,
        response,
    ):
        response_json = response.json()

        self.status = response_json.get("status")
        self.code = response_json.get("code")
        self.message = response_json.get("message")

        super(ClientError, self).__init__(self.message)

    def __str__(self):
        return self.message

    def __repr__(self):
        return f"{self.__class__.__name__}(status={self.status}, code={self.code}, message={self.message})"



from __future__ import annotations

import logging
from datetime import datetime
from typing import Callable

from requests.exceptions import RequestException

from licensespring.api.error import ClientError
from licensespring.licensefile.error import (
    ClockTamperedException,
    ConfigurationMismatch,
    ErrorType,
    LicenseStateException,
    TimeoutExpiredException,
    VMIsNotAllowedException,
)
from licensespring.licensefile.helpers import DateTimeHelper
from licensespring.licensefile.offline_activation_guard import OfflineActivation
from licensespring.watchdog import FeatureWatchdog, LicenseWatchdog


class License:
    """
    Represents and manages license operations including activation, deactivation,
    and consumption tracking for a software product.

    Methods:
        __init__(self, product, api_client, licensefile_handler): Initializes the license object with a product, API client, and licensefile handler.

        All Licenses:
            Getters:
                get_device_variables(self, get_from_be: bool = True): Retrieve device varaibles from server or from licensefile

                Licensefile (Offline):
                    start_date(self): Get the license start date
                    maintenance_period(self): Get the license maintenance period
                    is_maintence_period_expired(self): Check if maintence period has expired
                    maintenance_days_remaining(self): Remaining maintenance days

                    last_check(self): Get when was last license check
                    license_type(self): Get license type
                    max_activations(self): Get max activations
                    days_since_last_check(self): Number of days since last check
                    days_remaining(self): How many days remain until validity expires
                    max_transfers(self): Number of max transfers
                    is_device_transfer_allowed(self): If device transfer is allowed return True otherwise False
                    is_device_transfer_limited(self): Return True if there is limited for device transfer otherwise False
                    transfer_count(self): transfer count
                    customer_information(self): retrieve customer information
                    allow_unlimited_activations(self): If unlimited activations are allowed return True otherwise False
                    id(self): Get license id
                    license_enabled(self): If license enabled returns True otherwise False
                    license_active(self): If license is active returns True otherwise False
                    is_valid(self): Checks if license is enabled, active and non-expired then return True otherwise False (License not valid)
                    prevent_vm(self): if license prevents virtual machines return True otherwise False
                    is_trial(self): if license is trial return True otherwise False
                    check_license_status(self): Verifies the current status of the license, raising exceptions if the license is not enabled, not active, or expired.
                    local_check(self): Performs a local check of the license against product code, hardware ID, and VM restrictions.
                    is_grace_period(self, e: Exception): Determines if the current license state is within a grace period
                    get_grace_period(self): Get grace period value is hours.
                    custom_fields(self): Get list of custom fields
                    get_custom_field(self,field_name:str): Get custom fields value
                    get_product_details(self): Retrieve product details from licensefile
                    get_device_variable(self, variable_name: str): Retrieve device variable.
                Online:
                    check(self, include_expired_features=False, req_overages=-1): Performs an online check to sync license data with the backend.
                    product_details(self, include_latest_version: bool = False, include_custom_fields: bool = False): Retrieve product details from server.
            Post:

                Licensefile:
                    update_offline(self, path: str, reset_consumption: bool): Update License through refresh file.
                    set_device_variables(self, variables: dict, save: bool = True): Save device variables to licensefile.
                Online:
                    deactivate(self, delete_license=False): Deactivates the license and optionally deletes the local license file.
                    change_password(self, password: str, new_password: str): Change users password.
                    deactivate_offline(self, offline_path: str): Deactive offline license.
                    send_device_variables(self): Send device variables to LS server

        Time Limited:
            Getters:
                is_validity_period_expired(self): Determines whether the license's validity period has expired, based on the enabled state and validity period.
                validity_period(self): Get the license validity period
                is_expired(self): if validity period of license has expired returns True otherwise False

        Subscription:
            Getters:
                validity_with_grace_period(self): Get the validity period with grace period
                allow_grace_subscription_period(self): if grace period is allowed within subscription return True otherwise False
                is_subscription_grace_period_started(self): is grace period within subscription active
                subscription_grace_period(self): How many hours can subscription be in grace period

        Floating Licenses:
            Getters:
                is_floating_expired(self): Checks if a floating license has expired. Placeholder method that currently always returns False.
                expiry_date(self): Get expiry date of floating license
                is_borrowed(self): Check if license is borrowed
                borrow_until(self): get borrow end date
            Post:
                setup_license_watch_dog(self,callback,timeout):Initializes and starts the license watchdog with the specified callback and timeout.
                stop_license_watch_dog(self): Stops the license watchdog if it is currently running.
                floating_borrow(self,borrow_until:str): Attempts to borrow a floating license until the specified date.
                floating_release(self,throw_e:bool): Releases a borrowed floating license and updates the license status accordingly.

        Consumption Licenses:
            Getters:
                local_consumptions(self) : Get local consumptions
                max_consumptions(self): Get max consumptions
                total_consumptions(self): Get total consumptions
                max_overages(self): Get max overages
                allow_unlimited_consumptions(self): True if unlimited consumptions are allowed otherwise False
                consumption_reset(self): True if there is consumptions reset otherwise False
                allow_overages(self): True if overages are allowed otherwise False
                consumption_period(self): Consumption period
            Post:
                add_local_consumption(self, consumptions=1): Adds local consumption records for consumption-based licenses,
                sync_consumption(self, req_overages=-1): Syncs local consumption data with the server, adjusting for overages if specified.

        License Features:
            Getters:
                get_feature_data(self,feature_code:str): Get license feature fields dictionary.
            Post:
                Licensefile(offline):
                    add_local_feature_consumption(self, feature: str, consumptions=1): Add feature local consumption
                Online:
                    check_feature(self, feature: str, add_to_watchdog=False): Sync feature with LicenseSpring server or register floating feature
                    release_feature(self, feature: str): Release floating feature
                    sync_feature_consumption(self, feature): Sync consumption with the server
                    setup_feature_watch_dog(self,callback: Callable, timeout: int,deamon: bool =False): Setup feature watchdog for floating features
                    stop_feature_watch_dog(self): stop feature watchdog

    Attributes:
        product (str): The software product this license is associated with.
        api_client: An instance responsible for communicating with the licensing API.
        licensefile_handler: Handles local license file operations such as reading and writing.
    """

    def __init__(self, product, api_client, licensefile_handler) -> None:
        self.product = product
        self.api_client = api_client
        self.licensefile_handler = licensefile_handler
        self.watch_dog = None
        self.feature_watch_dog = None

    def is_floating_expired(self) -> bool:
        """
        Checks if floating license has expired

        Returns:
            bool: True if license has expired, otherwise False.
        """

        if self.licensefile_handler._cache.is_floating_license():
            return DateTimeHelper.has_time_started(self.expiry_date())
        return False

    def is_validity_period_expired(self) -> bool:
        """
        Determines whether the license's validity period has expired.

        Returns:
            bool: True if the validity period has expired or the license is disabled, False otherwise.
        """

        if self.is_expired():
            return True

        if isinstance(self.validity_with_grace_period(), datetime):
            if DateTimeHelper.has_time_started(self.validity_with_grace_period()):
                self.licensefile_handler._cache.set_boolean("is_expired", True)
                return True

        return False

    def validity_period(self) -> datetime:
        """
        Get validity period

        Returns:
            datetime: Validity period
        """
        return self.licensefile_handler._cache.get_validity_period()

    def validity_with_grace_period(self) -> datetime:
        """
        Get validity period

        Returns:
            datetime: Validity period
        """
        return self.licensefile_handler._cache.get_validity_with_grace_period()

    def maintenance_days_remaining(self):
        return DateTimeHelper.days_remain(self.maintenance_period())

    def days_remaining(self) -> int:
        return DateTimeHelper.days_remain(self.validity_with_grace_period())

    def customer_information(self) -> dict:
        return self.licensefile_handler._cache.get_customer()

    def id(self):
        return self.licensefile_handler._cache.get_id()

    def max_transfers(self) -> int:
        return self.licensefile_handler._cache.get_max_transfers()

    def transfer_count(self) -> int:
        return self.licensefile_handler._cache.get_transfer_count()

    def is_device_transfer_allowed(self) -> bool:
        return self.max_transfers() != -1

    def is_device_transfer_limited(self) -> bool:
        return self.max_transfers() != 0

    def days_since_last_check(self) -> int:
        return DateTimeHelper.days_since(self.last_check())

    def start_date(self) -> datetime:
        return self.licensefile_handler._cache.get_start_date()

    def maintenance_period(self) -> datetime:
        return self.licensefile_handler._cache.get_maintenance_period()

    def is_maintence_period_expired(self):
        return DateTimeHelper.has_time_started(self.maintenance_period())

    def last_check(self) -> datetime:
        return self.licensefile_handler._cache.get_last_check()

    def last_usage(self) -> datetime:
        return self.licensefile_handler._cache.get_last_usage()

    def license_type(self) -> str:
        return self.licensefile_handler._cache.get_license_type()

    def max_activations(self) -> int:
        return self.licensefile_handler._cache.get_max_activations()

    def metadata(self) -> dict:
        return self.licensefile_handler._cache.get_metadata()

    def allow_unlimited_activations(self) -> bool:
        return self.licensefile_handler._cache.get_allow_unlimited_activations()

    def allow_grace_subscription_period(self) -> bool:
        return self.licensefile_handler._cache.get_allow_grace_period()

    def is_subscription_grace_period_started(self) -> bool:
        if self.subscription_grace_period() == 0:
            return False

        return DateTimeHelper.has_time_started(self.validity_period())

    def get_grace_period(self) -> int:
        return self.licensefile_handler._cache.get_grace_period_conf()

    def subscription_grace_period(self) -> int:
        return self.licensefile_handler._cache.get_grace_period()

    def is_expired(self) -> bool:
        return self.licensefile_handler._cache.get_is_expired()

    def license_enabled(self) -> bool:
        return self.licensefile_handler._cache.get_license_enabled()

    def license_active(self) -> bool:
        return self.licensefile_handler._cache.get_license_active()

    def is_valid(self) -> bool:
        return (
            self.license_enabled() and self.license_active() and not self.is_expired()
        )

    def prevent_vm(self) -> bool:
        return self.licensefile_handler._cache.get_prevent_vm()

    def is_trial(self) -> bool:
        return self.licensefile_handler._cache.get_is_trial()

    # floating

    def expiry_date(self) -> datetime:
        return self.licensefile_handler._cache.get_floating_period()

    def borrow_until(self) -> datetime:
        return self.licensefile_handler._cache.get_borrowed_until()

    def is_borrowed(self) -> bool:
        return self.licensefile_handler._cache.get_is_borrowed()

    # consumption

    def local_consumptions(self) -> int:
        return self.licensefile_handler._cache.get_local_consumptions()

    def max_consumptions(self) -> int:
        return self.licensefile_handler._cache.get_max_consumptions()

    def total_consumptions(self) -> int:
        return self.licensefile_handler._cache.get_total_consumptions()

    def max_overages(self) -> int:
        return self.licensefile_handler._cache.get_max_overages()

    def allow_unlimited_consumptions(self) -> bool:
        return self.licensefile_handler._cache.get_allow_unlimited_consumptions()

    def consumption_reset(self) -> bool:
        return self.licensefile_handler._cache.get_consumption_reset()

    def allow_overages(self) -> bool:
        return self.licensefile_handler._cache.get_allow_overages()

    def consumption_period(self) -> str:
        return self.licensefile_handler._cache.get_consumption_period()

    # feature getters

    def get_feature_data(self, feature_code: str) -> dict:
        """
        Get feature dictionary

        Args:
            feature_code (str): feature code

        Returns:
            dict: feature dictionary
        """
        return self.licensefile_handler._cache.get_feature_dict(feature_code)

    def check_license_status(self) -> None:
        """
        Verifies the current status of the license, including its enablement, activation, and expiration.

        Raises:
            LicenseStateException: If the license is not enabled, not active, or expired.
        """

        if not self.license_enabled():
            raise LicenseStateException(
                ErrorType.LICENSE_NOT_ENABLED, "The license disabled"
            )

        if not self.license_active():
            raise LicenseStateException(
                ErrorType.LICENSE_NOT_ACTIVE, "The license is not active."
            )

        if self.is_validity_period_expired():
            raise LicenseStateException(
                ErrorType.LICENSE_EXPIRED, "The license is expired."
            )

    def check(self, include_expired_features=False) -> dict | None:
        """
        Performs an online license check, syncing the license data with the backend.
        This includes syncing consumptions for consumption-based licenses.

        Args:
            include_expired_features (bool, optional): If True, includes expired license features in the check.
                Defaults to False.
            req_overages (int, optional): Specifies the behavior for consumption overages.
                Use -1 to ignore, 0 to disable overages, and a positive value to enable overages
                up to the specified value. Defaults to -1.

        Returns:
            dict: The response from the license check operation.

        Raises:
            Exceptions: Various exceptions can be raised depending on the API client's implementation and the response from the licensing server.
        """

        try:
            # Add logic for floating server
            logging.info("Online check started")

            response = self.api_client.check_license(
                product=self.licensefile_handler._cache.product,
                bundle_code=getattr(
                    self.licensefile_handler._cache, "bundle_code", None
                ),
                hardware_id=None,
                license_key=getattr(
                    self.licensefile_handler._cache, "license_key", None
                ),
                license_id=getattr(self.licensefile_handler._cache, "id", None),
                username=getattr(self.licensefile_handler._cache, "username", None),
                include_expired_features=include_expired_features,
            )

            logging.info("Request send to server")

            self.licensefile_handler._cache.update_cache("check_license", response)
            self.licensefile_handler._cache.update_floating_period(
                self.licensefile_handler._cache.borrowed_until
            )

            logging.info("Updated licensefile with response")

            for (
                feature
            ) in self.licensefile_handler._cache.feature_manager.return_features_list():
                self.sync_feature_consumption(feature)

            logging.info("feature consumption synced")

            if self.licensefile_handler._cache.license_type == "consumption":
                self.sync_consumption(req_overages=-1)

            logging.info("license consumption synced")

            self.licensefile_handler._cache.reset_grace_period_start_date()

            logging.info("Online check successful")

            return response

        except ClientError as ex:
            self.licensefile_handler._cache.update_from_error_code(ex.code)

            logging.info(ex)
            raise ex

        except RequestException as ex:
            if (
                not self.licensefile_handler._cache.is_floating_license()
                or self.licensefile_handler._cache.is_active_floating_cloud()
            ) and self.is_grace_period(ex):
                return None

            raise RequestException("Grace period not allowed/passed")

        except Exception as ex:
            logging.info(ex)
            raise ex

        finally:
            self.licensefile_handler.save_licensefile()
            self.check_license_status()

    def deactivate(self, delete_license=False) -> None:
        """
        Deactivates the license and optionally deletes the local license file.

        Args:
            delete_license (bool, optional): If True, deletes the local license file upon deactivation.
                Defaults to False.
        """

        if not self.license_active():
            if delete_license:
                self.licensefile_handler.delete_licensefile()
            return None

        self.api_client.deactivate_license(
            product=self.licensefile_handler._cache.product,
            bundle_code=getattr(self.licensefile_handler._cache, "bundle_code", None),
            hardware_id=None,
            license_key=getattr(self.licensefile_handler._cache, "license_key", None),
            license_id=getattr(self.licensefile_handler._cache, "id", None),
            username=getattr(self.licensefile_handler._cache, "username", None),
        )

        if delete_license:
            self.licensefile_handler.delete_licensefile()
            return None

        self.licensefile_handler._cache.deactivate()
        self.licensefile_handler.save_licensefile()

    def local_check(self) -> None:
        """
        Performs a local check of the license, ensuring product code, hardware ID, VM (virtual machine), and other conditions are met.

        Raises:
            Various exceptions for different failure conditions.
        """
        try:
            if self.licensefile_handler._cache.product != self.product:
                raise ConfigurationMismatch(
                    ErrorType.PRODUCT_MISMATCH,
                    "License product code does not correspond to configuration product code",
                )

            if (
                self.licensefile_handler._cache.hardware_id
                != self.api_client.hardware_id_provider.get_id()
            ):
                raise ConfigurationMismatch(
                    ErrorType.HARDWARE_ID_MISMATCH,
                    "License hardware id does not correspond to configuration hardware id",
                )

            self.check_license_status()

            if self.api_client.hardware_id_provider.get_is_vm() != self.prevent_vm():
                raise VMIsNotAllowedException(
                    ErrorType.VM_NOT_ALLOWED, "Virtual machine not allowed."
                )

            if self.is_floating_expired():
                raise TimeoutExpiredException(
                    ErrorType.FLOATING_TIMEOUT, "Floating license timeout has expired."
                )

            if DateTimeHelper.has_time_expired(self.last_usage()):
                raise ClockTamperedException(
                    ErrorType.CLOCK_TAMPERED, "Detected cheating with local date time."
                )

        except LicenseStateException as e:
            self.licensefile_handler._cache.update_from_error_code(e.error_type.name)

            raise e

    def change_password(self, password: str, new_password: str) -> str:
        """
        Changes the password for user-based license.
        This method first checks the current license status to ensure it is active and not expired.
        It then attempts to change the password with the licensing server.

        Params:
            password (str): Old password of license user
            new_password (str): New password of license user

        Returns:
            str: password was changed.
        """

        self.check_license_status()

        response = self.api_client.change_password(
            username=getattr(self.licensefile_handler._cache, "username", None),
            password=password,
            new_password=new_password,
        )

        return response

    def add_local_consumption(self, consumptions=1) -> None:
        """
        Adds local consumption to the license
        Params:
            consumptions (int,optional): Number of consumptions.

        Returns: None
        """

        self.licensefile_handler._cache.update_consumption(consumptions)

    def add_local_feature_consumption(self, feature: str, consumptions=1) -> None:
        """
        Adds local consumption to the feature.
        Params:
            feature (str): feature code.
            consumptions (int,optional): Number of consumptions.
        """

        self.licensefile_handler._cache.update_feature_consumption(
            feature, consumptions
        )

    def sync_feature_consumption(self, feature) -> bool:
        """
        Syncs the local consumption data with the server for consumption-based licenses.

        Args:
            feature (str): feature code.

        Returns:
            bool: True if the consumption data was successfully synchronized; False otherwise.

        """

        if not hasattr(self.licensefile_handler._cache.feature_manager, feature):
            return False

        feature_obj = getattr(self.licensefile_handler._cache.feature_manager, feature)

        if not hasattr(feature_obj, "local_consumption"):
            return False

        if (
            self.licensefile_handler._cache.get_feature_object_field(
                feature_obj, "local_consumption"
            )
            == 0
        ):
            return False

        try:
            response = self.api_client.add_feature_consumption(
                product=self.product,
                bundle_code=getattr(
                    self.licensefile_handler._cache, "bundle_code", None
                ),
                hardware_id=None,
                license_key=getattr(
                    self.licensefile_handler._cache, "license_key", None
                ),
                license_id=getattr(self.licensefile_handler._cache, "id", None),
                username=getattr(self.licensefile_handler._cache, "username", None),
                feature=feature,
                consumptions=self.licensefile_handler._cache.get_feature_object_field(
                    feature_obj, "local_consumption"
                ),
            )

        except ClientError as ex:
            logging.info(ex)
            raise ex

        except RequestException as ex:
            if self.is_grace_period(ex):
                self.licensefile_handler.save_licensefile()
                self.check_license_status()

                return False

            raise RequestException("Grace period not allowed/passed")

        except Exception as ex:
            logging.info(ex)
            raise ex

        else:
            self.licensefile_handler._cache.update_cache(
                "feature_consumption", response, feature
            )

            self.licensefile_handler._cache.reset_grace_period_start_date()

            self.licensefile_handler.save_licensefile()

            return True

    def sync_consumption(self, req_overages=-1) -> bool:
        """
        Syncs the local consumption data with the server for consumption-based licenses.

        Args:
            req_overages (int, optional): Specifies the behavior for requesting consumption overages.
                Defaults to -1, which means no overage request is made. A value of 0 disables overages,
                and a positive value requests permission for overages up to the specified value.

        Returns:
            bool: True if the consumption data was successfully synchronized; False otherwise.

        Side Effects:
            Resets local consumption count after successful synchronization.
        """

        if not hasattr(self.licensefile_handler._cache, "local_consumption"):
            return False

        if self.local_consumptions() == 0 and req_overages < 0:
            return False

        try:
            if req_overages == 0:
                max_overages = req_overages
                allow_overages = False

            elif req_overages > 0:
                max_overages = req_overages
                allow_overages = True

            else:
                max_overages = None
                allow_overages = None

            response = self.api_client.add_consumption(
                product=self.licensefile_handler._cache.product,
                bundle_code=getattr(
                    self.licensefile_handler._cache, "bundle_code", None
                ),
                license_key=getattr(
                    self.licensefile_handler._cache, "license_key", None
                ),
                license_id=getattr(self.licensefile_handler._cache, "id", None),
                username=getattr(self.licensefile_handler._cache, "username", None),
                consumptions=self.local_consumptions(),
                max_overages=max_overages,
                allow_overages=allow_overages,
            )

        except ClientError as ex:
            logging.info(ex)
            raise ex

        except RequestException as ex:
            if self.is_grace_period(ex):
                self.licensefile_handler.save_licensefile()
                self.check_license_status()

                return False

            raise RequestException("Grace period not allowed/passed")

        except Exception as ex:
            logging.info(ex)
            raise ex

        else:
            self.licensefile_handler._cache.update_cache(
                "license_consumption", response
            )

            self.licensefile_handler._cache.reset_grace_period_start_date()
            self.licensefile_handler.save_licensefile()

            return True

    def is_grace_period(self, ex: Exception) -> bool:
        """
        Determines if the license is currently within its grace period following a specific exception.
        The grace period logic is activated only if the license cache has a 'grace_period' attribute set,
        and the passed exception is of type 'RequestException', typically indicating a communication
        error with the licensing server.

        Returns:
            bool: True if the license is within its grace period, False otherwise.

        Side Effects:
            - If the license is within its grace period and a 'RequestException' occurs, this method
            updates the grace period start date in the license cache to the current time.
        """

        if not hasattr(self.licensefile_handler._cache, "grace_period_conf"):
            return False

        elif self.licensefile_handler._cache.grace_period_conf > 0 and isinstance(
            ex, RequestException
        ):
            self.licensefile_handler._cache.update_grace_period_start_date()

            return DateTimeHelper.has_time_expired(
                self.licensefile_handler._cache.grace_period_end_date()
            )

        return False

    def setup_license_watch_dog(
        self,
        callback: Callable,
        timeout: int,
        run_immediately: bool = True,
        deamon: bool = False,
    ) -> None:
        """
        Initializes and starts the license watchdog with the specified callback and timeout.

        Args:
            callback: A callable to be executed by the watchdog in response to specific events or conditions.
            timeout: The period in minutes after which the watchdog should perform its checks.
            deamon: run thread as deamon
            run_immediately: run license check immediately, if False wait for timeout first.

        Side Effects:
            - Instantiates the LicenseWatchdog class and stores the instance.
            - Starts the watchdog thread.
        """

        self.watch_dog = LicenseWatchdog(self, callback, timeout)
        self.watch_dog.run(run_immediately=run_immediately, deamon=deamon)

    def stop_license_watch_dog(self) -> None:
        """
        Stops the license watchdog if it is currently running.

        Side Effects:
            - Stops the watchdog thread, if it exists.
        """

        if self.watch_dog:
            self.watch_dog.stop()

    def setup_feature_watch_dog(
        self, callback: Callable, timeout: int, deamon: bool = False
    ):
        """
        Initializes and starts the feature watchdog with the specified callback and timeout.

        Args:
            callback (Callable): A callable to be executed by the watchdog in response to specific events or conditions.
            timeout (int): The period in minutes after which the watchdog should perform its checks.
            deamon (bool, optional): Run thread as deamon. Defaults to False.
        """
        logging.info("Setting up feature watchdog")
        self.feature_watch_dog = FeatureWatchdog(self, callback, timeout)

        for (
            feature
        ) in self.licensefile_handler._cache.feature_manager.attributes_to_list():
            self.feature_watch_dog.add_feature(feature["code"])

        self.feature_watch_dog.run(deamon=deamon)

    def stop_feature_watch_dog(self):
        """
        Stops the license watchdog if it is currently running.

        Side Effects:
            - Stops the watchdog thread, if it exists.
        """
        if self.feature_watch_dog != None:
            self.feature_watch_dog.stop()

    def floating_borrow(self, borrow_until: str, password: str = None) -> None:
        """
        Attempts to borrow a floating license until the specified date.

        Args:
            borrow_until: A string representing the date until which the license should be borrowed.

        Returns:
            None

        Side Effects:
            - Checks if the license is floating or floating cloud type and attempts to borrow it.
            - Updates the license cache and stops the license watchdog if borrowing is successful.
        """

        if not self.licensefile_handler._cache.is_floating_license():
            return None

        response = self.api_client.floating_borrow(
            product=self.licensefile_handler._cache.product,
            bundle_code=getattr(self.licensefile_handler._cache, "bundle_code", None),
            hardware_id=None,
            license_key=getattr(self.licensefile_handler._cache, "license_key", None),
            license_id=getattr(self.licensefile_handler._cache, "id", None),
            username=getattr(self.licensefile_handler._cache, "username", None),
            password=password,
            borrowed_until=borrow_until,
        )

        self.licensefile_handler._cache.set_boolean("is_borrowed", True)
        self.stop_license_watch_dog()
        self.licensefile_handler._cache.update_cache("normal", response)
        self.licensefile_handler._cache.update_floating_period(
            self.licensefile_handler._cache.borrowed_until
        )
        self.licensefile_handler.save_licensefile()

    def floating_release(self, throw_e: bool):
        """
        Releases a borrowed floating license and updates the license status accordingly.

        Args:
            throw_e: A boolean indicating whether to raise an exception on failure.

        Returns:
            None

        Side Effects:
            - Attempts to release the floating license and update the license cache.
            - Logs and potentially raises an exception if an error occurs during release.
        """

        if not self.licensefile_handler._cache.is_floating_license():
            return None

        try:
            self.check_license_status()

            # Add logic for both floating cloud and floating license
            if self.licensefile_handler._cache.is_active_floating_cloud():
                self.api_client.floating_release(
                    product=self.licensefile_handler._cache.product,
                    bundle_code=getattr(
                        self.licensefile_handler._cache, "bundle_code", None
                    ),
                    hardware_id=None,
                    license_key=getattr(
                        self.licensefile_handler._cache, "license_key", None
                    ),
                    license_id=getattr(self.licensefile_handler._cache, "id", None),
                    username=getattr(self.licensefile_handler._cache, "username", None),
                )

                self.licensefile_handler._cache.release_license()
                self.licensefile_handler.save_licensefile()

        except Exception as ex:
            if throw_e:
                logging.info(ex)
                raise ex

    def check_feature(self, feature: str, add_to_watchdog=False) -> None:
        """
        Checks for a specific license feature and updates the license cache accordingly.

        Args:
            feature: feature code.
            add_to_watchdog: A boolean indicating whether to add the feature check to a watchdog routine.

        Returns:
            None
        """
        try:
            response = self.api_client.check_license_feature(
                product=self.licensefile_handler._cache.product,
                feature=feature,
                bundle_code=getattr(
                    self.licensefile_handler._cache, "bundle_code", None
                ),
                hardware_id=None,
                license_key=getattr(
                    self.licensefile_handler._cache, "license_key", None
                ),
                license_id=getattr(self.licensefile_handler._cache, "id", None),
                username=getattr(self.licensefile_handler._cache, "username", None),
            )

            self.licensefile_handler._cache.register_feature(feature)
            self.licensefile_handler._cache.update_cache(
                "register_feature", response, feature
            )

            if self.feature_watch_dog != None and add_to_watchdog:
                self.feature_watch_dog.add_feature(feature)

            self.licensefile_handler.save_licensefile()

        except Exception as ex:
            logging.info(ex)
            raise ex

    def release_feature(self, feature: str):
        """
        Releases a borrowed license feature and updates the license cache accordingly.

        Args:
            feature: The feature code.

        Returns:
            None
        """
        try:
            self.api_client.floating_feature_release(
                product=self.licensefile_handler._cache.product,
                feature=feature,
                bundle_code=getattr(
                    self.licensefile_handler._cache, "bundle_code", None
                ),
                hardware_id=None,
                license_key=getattr(
                    self.licensefile_handler._cache, "license_key", None
                ),
                license_id=getattr(self.licensefile_handler._cache, "id", None),
                username=getattr(self.licensefile_handler._cache, "username", None),
            )

            self.licensefile_handler._cache.release_feature(feature)
            if self.feature_watch_dog != None:
                self.feature_watch_dog.remove_feature(feature)

            self.licensefile_handler.save_licensefile()

        except Exception as ex:
            logging.info(ex)
            raise ex

    def update_offline(self, path: str, reset_consumption: bool) -> bool:
        """
        Updates license via refresh file

        Args:
            path (str): path of the refresh file
            reset_consumption (bool): True resets consumption otherwise False

        Raises:
            ConfigurationMismatch: The update file does not belong to this device
            ConfigurationMismatch: The update file does not belong to this product

        Returns:
            bool: True if license was successfully updated otherwise False
        """

        data = self.licensefile_handler.load_offline_response(path)
        decoded_data = self.api_client.check_offline_load(data)

        if decoded_data["hardware_id"] != self.licensefile_handler._cache.hardware_id:
            raise ConfigurationMismatch(
                ErrorType.HARDWARE_ID_MISMATCH,
                " The update file does not belong to this device. ",
            )

        if (
            decoded_data["product_details"]["short_code"]
            != self.licensefile_handler._cache.product
        ):
            raise ConfigurationMismatch(
                ErrorType.PRODUCT_MISMATCH,
                " The update file does not belong to this product.",
            )

        if (
            reset_consumption == True
            and self.licensefile_handler._cache.license_type == "consumption"
        ):
            self.licensefile_handler._cache.reset_local_consumption()

        self.licensefile_handler._cache.update_cache(
            "update_license_offline", decoded_data
        )
        self.licensefile_handler.save_licensefile()

        return True

    def deactivate_offline(self, offline_path: str) -> str:
        """
        Generates .req file for the offline deactivation

        Args:
            offline_path (str): path of the .req file

        Returns:
            str: path of the deactivation file
        """

        self.licensefile_handler._cache.deactivate()
        data = self.api_client.deactivate_offline_dump(
            product=self.product,
            bundle_code=getattr(self.licensefile_handler._cache, "bundle_code", None),
            hardware_id=None,
            license_key=getattr(self.licensefile_handler._cache, "license_key", None),
            license_id=getattr(self.licensefile_handler._cache, "id", None),
            username=getattr(self.licensefile_handler._cache, "username", None),
        )
        offline_data = OfflineActivation()
        offline_data.set_is_activation(False)
        offline_data.set_data(data)

        return self.licensefile_handler.create_request_file(offline_data, offline_path)

    def product_details(
        self, include_latest_version: bool = False, include_custom_fields: bool = False
    ) -> dict:
        """
        Update product details from LicenseSpring server

        Args:
            include_latest_version (bool, optional): Lateset version information. Defaults to False.
            include_custom_fields (bool, optional): custom fields information. Defaults to False.

        Returns:
            dict: response
        """
        response = self.api_client.product_details(
            self.product, include_latest_version, include_custom_fields
        )

        self.licensefile_handler._cache.update_cache("product_details", response)
        self.licensefile_handler.save_licensefile()

        return response

    def get_product_details(self) -> dict:
        """
        Retrieve product details from licensefile

        Returns:
            dict: Product details
        """
        return self.licensefile_handler._cache.get_product_details()

    def set_device_variables(self, variables: dict, save: bool = True):
        """
        Set device variables locally

        Args:
            variables (dict): variables dict
            save (bool, optional): Save cache to licensefile. Defaults to True.
        """
        self.licensefile_handler._cache.set_variables(variables)
        if save:
            self.licensefile_handler.save_licensefile()

    def send_device_variables(self) -> bool:
        """
        Send device variables to LicenseSpring server. Handles GracePeriod

        Raises:
            ex: RequestException (Grace period not allowed)

        Returns:
            bool: True if new variables are sent to LicenseSpring server otherwise, False
        """
        try:
            new_varaibles = (
                self.licensefile_handler._cache.get_device_variables_for_send()
            )

            if len(new_varaibles) == 0:
                return False

            self.api_client.track_device_variables(
                product=self.product,
                bundle_code=getattr(
                    self.licensefile_handler._cache, "bundle_code", None
                ),
                hardware_id=None,
                license_key=getattr(
                    self.licensefile_handler._cache, "license_key", None
                ),
                license_id=getattr(self.licensefile_handler._cache, "id", None),
                username=getattr(self.licensefile_handler._cache, "username", None),
                variables=new_varaibles,
            )

            return True

        except RequestException as ex:
            if self.is_grace_period(ex):
                return False

            logging.info(ex)
            raise ex

    def get_device_variable(self, variable_name: str) -> dict:
        """
        Get device variable if exists

        Args:
            variable_name (str): variable name

        Returns:
            dict: variable dictionary
        """
        return self.licensefile_handler._cache.get_variable(variable_name)

    def get_device_variables(self, get_from_be: bool = True) -> list:
        """
        Get device variables from server or locally

        Args:
            get_from_be (bool, optional): If True collects data from LicenseSpring server. Defaults to True.

        Raises:
            ex: RequestException (Grace period not allowed)

        Returns:
            list: List of device variables
        """
        if get_from_be:
            try:
                response = self.api_client.get_device_variables(
                    product=self.product,
                    bundle_code=getattr(
                        self.licensefile_handler._cache, "bundle_code", None
                    ),
                    hardware_id=None,
                    license_key=getattr(
                        self.licensefile_handler._cache, "license_key", None
                    ),
                    license_id=getattr(self.licensefile_handler._cache, "id", None),
                    username=getattr(self.licensefile_handler._cache, "username", None),
                )

            except RequestException as ex:
                if not self.is_grace_period():
                    logging.info(ex)
                    raise ex

            self.licensefile_handler._cache.update_cache("device_variables", response)
            self.licensefile_handler.save_licensefile()

        return self.licensefile_handler._cache.get_variables()

    def custom_fields(self) -> list:
        """
        Get custom fields

        Returns:
            list: list of custom fields -> [{name, data_type, value},..]
        """
        return self.licensefile_handler._cache.get_custom_fields()

    def get_custom_field(self, field_name: str) -> str:
        """
        Get custom field value

        Args:
            field_name (str): custom field name

        Returns:
            str: custom field value
        """
        return self.licensefile_handler._cache.get_custom_field(field_name)



from __future__ import annotations

import json
import threading
from datetime import datetime, timedelta, timezone

import licensespring
from licensespring.licensefile.device_variables import VariablesManager
from licensespring.licensefile.error import (
    ConsumptionError,
    ErrorType,
    LicenseSpringTypeError,
)
from licensespring.licensefile.feature_manager import FeatureManager
from licensespring.licensefile.product_details import ProductDetailsManager


class LicenseData:

    """
    A class for handling License file fields, providing functionalities for managing license data.

    Attributes:
        grace_period_conf (int): Defines how many hours can license be in offline mode.
        product (str): Identifier for the software product associated with this license.
        hardware_id (str): Unique hardware ID of the device using the license, used for license binding.
        version (str): Version of the Python SDK, fetched from licensespring.version.
        license_key (str, optional): The key used for license activation. Default is None.
        floating_period (int, optional): Defines the period for a floating license. Default is None.
        last_check (datetime, optional): Timestamp of the last license check. Default is None.
        last_usage (datetime, optional): Timestamp of the last usage of the license. Default is None.

    Methods:
        features_setup(): Sets up local consumption count for consumption-based features.
        local_consumption_setup(): Initializes local consumption count for licenses of type 'consumption'.
        to_json(): Serializes the license data to a JSON string.
        from_json_to_attr(licensefile_dict): Deserializes JSON string or dictionary to license data attributes.
        update_cache(method, response): Updates the license data in memory based on the response from a license operation.

        validity_with_grace_period(): Calculates and returns the expiration date.

        is_grace_period_started(): If grace period started returns True, otherwise False.
        grace_period_end_date(): Calculates end date of grace period.
        update_grace_period_start_date(): Sets grace_period_start_date to UTC.
        reset_grace_period_start_date(): Resets grace period start date.

        update_from_error_code(error_code): Updates the state of the object based on the provided error code.

        is_active_floating_cloud(): Determines whether the license is active floating cloud.

        deactivate(): Deactivates the current license and updates times_activated and license_active.

        update_password(): updates password.

        update_consumption(consumptions:int):Adds local consumptions for consumption-based licenses.
        update_feature_consumption(feature, consumptions=1): Adds a specified number of local consumptions to a feature.

        release_license():Handles operations on licensefile for releasing license.
        register_feature(feature_code): Handles feature check.
        release_feature(feature_code:str): Handles feature release.
        set_boolean(field_name:str,bool:bool): set boolean of field_name.

    """

    datetime_attributes = [
        "last_check",
        "last_usage",
        "floating_period",
        "maintenance_period",
        "start_date",
        "validity_period",
        "borrowed_until",
        "grace_period_start_date",
    ]

    def __init__(self, product, hardwareID, grace_period_conf):
        self.grace_period_conf = grace_period_conf
        self.product = product
        self.hardware_id = hardwareID
        self.version = licensespring.version
        self.feature_manager = FeatureManager()
        self.device_var_manager = VariablesManager()
        self.product_details_manager = ProductDetailsManager()

        self.license_key = None
        self.floating_period = datetime.now(timezone.utc).replace(tzinfo=None)

        self.last_check = None
        self.last_usage = None

        self._lock = threading.Lock()

    def local_consumption_setup(self):
        """
        Initializes the local consumption attribute for licenses of type 'consumption' if it's not already set.
        This method assumes the existence of a license_type attribute to check against 'consumption'.
        """

        if (
            not hasattr(self, "local_consumption")
            and self.license_type == "consumption"
        ):
            self.local_consumption = 0

    def grace_period_setup(self):
        if not hasattr(self, "grace_period_start_date"):
            self.grace_period_start_date = datetime(year=2099, month=1, day=1)

    def to_json(self) -> json:
        """
        Serializes the license data attributes to a JSON string, converting datetime objects to ISO format.

        Returns:
            str: The serialized JSON string of the license data.
        """
        with self._lock:
            data = {}
            for key, value in self.__dict__.items():
                if isinstance(value, datetime):
                    data[key] = value.isoformat()

                elif key == "feature_manager":
                    data["product_features"] = self.feature_manager.attributes_to_list()

                elif key == "device_var_manager":
                    data[
                        "device_variables"
                    ] = self.device_var_manager.attributes_to_list()

                elif key == "product_details_manager":
                    data["product_details"] = self.product_details_manager.__dict__

                elif key == "_lock":
                    pass

                else:
                    data[key] = value

            return json.dumps(data)

    def from_json_to_attr(self, licensefile_dict: dict):
        """
        Deserializes a JSON string or dictionary to update the license data attributes, converting ISO format strings back to datetime objects for specific fields.

        Args:
            licensefile_dict (dict): The dictionary containing license data to be deserialized.
        """

        for key, value in licensefile_dict.items():
            if key in LicenseData.datetime_attributes and value != None:
                if value.endswith("Z"):
                    value = value[:-1]
                setattr(self, key, datetime.fromisoformat(value).replace(tzinfo=None))

            elif key == "product_details":
                self.product_details_manager.json_to_attribute(
                    licensefile_dict["product_details"]
                )

            elif key == "product_features":
                self.feature_manager.json_to_attribute(
                    licensefile_dict["product_features"]
                )

            else:
                setattr(self, key, value)

    def update_cache(self, method: str, response: dict, feature: str = None):
        """
        Updates licensefile data inside memory.

        Parameters:
        response (dict): Response of license activation or license check.

        Returns: None
        """
        with self._lock:
            self.last_usage = datetime.now(timezone.utc).replace(tzinfo=None)

            if method == "check_license":
                self.last_check = datetime.now(timezone.utc).replace(tzinfo=None)
                self.from_json_to_attr(response)

            elif method == "activate_license":
                self.from_json_to_attr(response)
                self.local_consumption_setup()
                self.grace_period_setup()

                self.license_enabled = True
                self.is_expired = False
                self.license_active = True

            elif method == "activate_license_offline":
                response["prevent_vm"] = response["product_details"]["prevent_vm"]
                self.from_json_to_attr(response)
                self.local_consumption_setup()
                self.grace_period_setup()

                self.license_enabled = True
                self.is_expired = False
                self.license_active = True

            elif method == "update_license_offline":
                self.from_json_to_attr(response)

            elif method == "product_details":
                self.product_details_manager.json_to_attribute(response)

                if "custom_fields" in response:
                    self.custom_fields = response["custom_fields"]

                if "latest_version" in response:
                    self.latest_version = response["latest_version"]

            elif method == "device_variables":
                self.device_var_manager.json_to_attribute(response)

            elif method == "normal":
                self.from_json_to_attr(response)

            elif method == "license_consumption":
                self.from_json_to_attr(response)
                self.local_consumption = 0

            elif method == "feature_consumption":
                feature_obj = getattr(self.feature_manager, feature)
                feature_obj.json_to_attribute(response)
                feature_obj.local_consumption = 0

            elif method == "register_feature":
                feature_obj = getattr(self.feature_manager, feature)
                feature_obj.json_to_attribute(response)

    def get_validity_period(self):
        with self._lock:
            return self.validity_period

    def get_validity_with_grace_period(self) -> datetime | None:
        """
        Calculates and returns the expiration date and time of an object with an optional grace period extension.
        """

        with self._lock:
            if hasattr(self, "grace_period") and self.validity_period != None:
                if self.grace_period > 0:
                    return self.validity_period + timedelta(hours=self.grace_period)

            return self.validity_period

    def is_grace_period_started(self) -> bool:
        return self.grace_period_start_date != datetime(year=2099, month=1, day=1)

    def grace_period_end_date(self) -> datetime:
        return self.grace_period_start_date + timedelta(hours=self.grace_period_conf)

    def update_grace_period_start_date(self) -> None:
        with self._lock:
            if not self.is_grace_period_started():
                self.grace_period_start_date = datetime.now(timezone.utc).replace(
                    tzinfo=None
                )

    def reset_grace_period_start_date(self) -> None:
        with self._lock:
            if self.is_grace_period_started():
                self.grace_period_start_date = datetime(year=2099, month=1, day=1)

    def update_from_error_code(self, error_code: str):
        """
        Updates the state of the object based on the provided error code.

        Args:
            error_code (str): A string representing the error code which dictates how the object's state should be updated. The method recognizes three specific error codes:
                - "LICENSE_NOT_ACTIVE": Indicates the license is not currently active. Sets `self.license_active` to False.
                - "LICENSE_NOT_ENABLED": Indicates the license is not enabled. Sets `self.license_enabled` to False.
                - "LICENSE_EXPIRED": Indicates the license has expired. Sets `self.is_expired` to True.

        The method converts `error_code` to uppercase to ensure case-insensitive comparison.
        """

        with self._lock:
            if error_code.upper() == "LICENSE_NOT_ACTIVE":
                self.license_active = False

            elif error_code.upper() == "LICENSE_NOT_ENABLED":
                self.license_enabled = False

            elif error_code.upper() == "LICENSE_EXPIRED":
                self.is_expired = True

    def is_active_floating_cloud(self):
        """
        Determines whether the license is active floating cloud

        Returns:
            bool: True if the floating cloud is active, False otherwise.
        """
        with self._lock:
            if self.is_floating_cloud:
                if self.floating_period > datetime.now(timezone.utc).replace(
                    tzinfo=None
                ):
                    return True

            return False

    def deactivate(self):
        """
        Deactivates the current license and updates times_activated and license_active.

        1. Decrements the `times_activated` counter if it is greater than 0. This counter tracks the number of times the license has been activated. The decrement operation signifies one less active instance of the license.

        2. If `times_activated` is already at 0, it leaves it unchanged. This prevents the counter from going negative, which would be semantically incorrect.

        3. Sets `self.license_active` to False, indicating that the license is no longer active.

        This method is typically called when the license needs to be deactivated.

        """

        with self._lock:
            if self.times_activated > 0:
                self.times_activated -= 1

            else:
                self.times_activated = 0

            self.license_active = False

    def update_consumption(self, consumptions: int):
        """
        Adds local consumptions for consumption-based licenses.

        Args:
            consumptions (int): The number of consumptions to add locally.

        Returns:
            dict: The updated license cache reflecting the new consumption count.

        Raises:
            LicenseTypeError: If the license is not of type 'consumption'.
            ConsumptionError: If adding the consumptions would exceed the allowed maximum.
        """

        with self._lock:
            if self.license_type != "consumption":
                raise LicenseSpringTypeError(
                    ErrorType.WRONG_LICENSE_TYPE,
                    f" WRONG License Type: {self.license_type}",
                )
            elif self.allow_unlimited_consumptions:
                self.local_consumption += consumptions

            elif (
                self.max_consumptions + self.max_overages
                < self.local_consumption + consumptions + self.total_consumptions
            ):
                raise ConsumptionError(
                    ErrorType.NOT_ENOUGH_LICENSE_CONSUMPTIONS,
                    "Not enough conusmptions left!",
                )

            else:
                self.local_consumption += consumptions

    def reset_local_consumption(self):
        with self._lock:
            self.local_consumption = 0

    def is_floating_license(self):
        return self.is_floating or self.is_floating_cloud

    def update_feature_consumption(self, feature, consumptions=1):
        """
        Adds a specified number of consumptions to a local feature, updating its consumption count.
        """

        with self._lock:
            self.feature_manager.add_local_consumption(feature, consumptions)

    def update_floating_period(self, end_date: datetime) -> None:
        """
        Updates floating end time if needed

        Args:
            end_date (datetime): end date of floating
        """

        with self._lock:
            if self.is_floating_license():
                floating_end = (
                    timedelta(minutes=self.floating_timeout) + self.floating_period
                )

                if isinstance(end_date, datetime):
                    if end_date != floating_end:
                        floating_end = end_date

                self.floating_period = floating_end

    def release_license(self):
        """
        Handles operations on licensefile for releasing license

        1. Sets `self.is_borrowed` to False, indicating the license is no longer borrowed.
        2. Resets `self.borrowed_until` to None, clearing the timestamp until which the license was borrowed.
        3. Decrements `self.floating_in_use_devices` by 1, reducing the count of devices currently using a floating license by one.
        """

        with self._lock:
            self.is_borrowed = False
            self.borrowed_until = None
            self.floating_in_use_devices -= 1
            self.floating_period = datetime.now(timezone.utc).replace(tzinfo=None)

    def register_feature(self, feature_code: str):
        """
        Handles feature check
        """

        with self._lock:
            self.feature_manager.register_feature(feature_code)

    def release_feature(self, feature_code: str):
        """
        Handles feature release
        """

        with self._lock:
            self.feature_manager.release_feature(feature_code)

    def set_boolean(self, field_name: str, bool: bool):
        """
        Changes/Sets LicenseData field to boolean value

        Args:
            field_name (str): field name
            bool (bool): True/False
        """

        with self._lock:
            setattr(self, field_name, bool)

    def get_feature_object(self, feature: str):
        with self._lock:
            return getattr(self.feature_manager, feature, None)

    def get_feature_object_field(self, feature_object: object, field: str):
        with self._lock:
            return getattr(feature_object, field)

    def get_feature_dict(self, feature_code):
        with self._lock:
            return self.feature_manager.get_feature_dict(feature_code)

    def get_variable(self, variable_name: str):
        with self._lock:
            return self.device_var_manager.get_device_variable(variable_name)

    def set_variables(self, variables: dict):
        with self._lock:
            self.device_var_manager.set_variables(variables)

    def get_variables(self):
        with self._lock:
            return self.device_var_manager.attributes_to_list()

    def get_device_variables_for_send(self) -> dict:
        with self._lock:
            return self.device_var_manager.get_device_variable_for_send()

    def get_custom_fields(self):
        with self._lock:
            return self.custom_fields

    def get_custom_field(self, field_name: str):
        for field in self.custom_fields:
            if field["name"] == field_name:
                return field["value"]

    def get_customer(self) -> dict:
        with self._lock:
            return self.customer

    def get_product_details(self):
        return self.product_details_manager.__dict__

    def get_is_borrowed(self):
        with self._lock:
            return self.is_borrowed

    def get_borrowed_until(self):
        with self._lock:
            return self.borrowed_until

    def get_floating_period(self):
        with self._lock:
            return self.floating_period

    def get_max_transfers(self):
        with self._lock:
            return self.max_transfers

    def get_transfer_count(self):
        with self._lock:
            return self.transfer_count

    def get_id(self):
        with self._lock:
            return self.id

    def get_start_date(self):
        with self._lock:
            return self.start_date

    def get_maintenance_period(self):
        with self._lock:
            return self.maintenance_period

    def get_last_check(self):
        with self._lock:
            return self.last_check

    def get_last_usage(self):
        with self._lock:
            return self.last_usage

    def get_max_activations(self):
        with self._lock:
            return self.max_activations

    def get_allow_unlimited_activations(self):
        with self._lock:
            return self.allow_unlimited_activations

    def get_is_trial(self):
        with self._lock:
            return self.is_trial

    def get_allow_grace_period(self):
        with self._lock:
            return self.allow_grace_period

    def get_grace_period_conf(self):
        with self._lock:
            return self.grace_period_conf

    def get_grace_period(self):
        with self._lock:
            return self.grace_period

    def get_is_expired(self):
        with self._lock:
            return self.is_expired

    def get_license_active(self):
        with self._lock:
            return self.license_active

    def get_license_enabled(self):
        with self._lock:
            return self.license_enabled

    def get_license_type(self):
        with self._lock:
            return self.license_type

    def get_prevent_vm(self):
        with self._lock:
            return self.prevent_vm

    def get_metadata(self):
        with self._lock:
            return self.metadata

    def get_local_consumptions(self) -> int:
        with self._lock:
            return getattr(self, "local_consumption", None)

    def get_max_consumptions(self) -> int:
        with self._lock:
            return getattr(self, "max_consumptions", None)

    def get_total_consumptions(self) -> int:
        with self._lock:
            return getattr(self, "total_consumptions", None)

    def get_max_overages(self) -> int:
        with self._lock:
            return getattr(self, "max_overages", None)

    def get_consumption_reset(self) -> bool:
        with self._lock:
            return getattr(self, "reset_consumption", None)

    def get_allow_unlimited_consumptions(self) -> bool:
        with self._lock:
            return getattr(self, "allow_unlimited_consumptions", None)

    def get_consumption_period(self) -> str:
        with self._lock:
            return getattr(self, "consumption_period", None)

    def get_allow_overages(self) -> bool:
        with self._lock:
            return getattr(self, "allow_overages", None)



from datetime import datetime, timezone


class DateTimeHelper:
    @classmethod
    def days_remain(cls, input_datetime: datetime) -> int:
        """
        Checks how many days are left from current UTC time

        Args:
            input_datetime (datetime): value which is compared with UTC time

        Returns:
            int: days left
        """
        if not isinstance(input_datetime, datetime):
            return None

        diff = input_datetime - datetime.now(timezone.utc).replace(tzinfo=None)
        return diff.days

    @classmethod
    def days_since(cls, input_datetime: datetime) -> int:
        """
        Checks how many days are left from current UTC time

        Args:
            input_datetime (datetime): value which is compared with UTC time

        Returns:
            int: days left
        """
        if not isinstance(input_datetime, datetime):
            return None

        diff = datetime.now(timezone.utc).replace(tzinfo=None) - input_datetime
        return diff.days

    @classmethod
    def has_time_expired(cls, input_datetime: datetime) -> bool:
        """
        Checks if time has expired

        Args:
            input_datetime (datetime): value which is compared with UTC time

        Returns:
            bool: Returns True if time has expired otherwise False
        """
        if not isinstance(input_datetime, datetime):
            return True
        return input_datetime > datetime.now(timezone.utc).replace(tzinfo=None)

    @classmethod
    def has_time_started(cls, input_datetime: datetime) -> bool:
        """
        Checks if time has started

        Args:
            input_datetime (datetime): value which is compared with UTC time

        Returns:
            bool: Returns True if time has started otherwise False
        """
        if not isinstance(input_datetime, datetime):
            return True
        return input_datetime < datetime.now(timezone.utc).replace(tzinfo=None)



from __future__ import annotations

import base64
import json
import logging
import os
from datetime import datetime, timezone

from licensespring.licensefile.config import Configuration
from licensespring.licensefile.default_crypto import DefaultCryptoProvider
from licensespring.licensefile.error import (
    ErrorType,
    LicenseActivationException,
    LicenseDeleted,
)
from licensespring.licensefile.license_data import LicenseData
from licensespring.licensefile.offline_activation_guard import (
    OfflineActivation,
    OfflineActivationGuard,
)


class DataStorage(DefaultCryptoProvider):
    """
    Extends DefaultCryptoProvider to handle license file cache operations, including saving to and loading from a .key file.

    Attributes:
        _conf (Configuration): Configuration object containing settings and parameters.
        _cache (LicenseData): Instance of LicenseData for managing license attributes in memory.
        _filename (str): The name of the file used for storing the license information.
        _license_path (str): The file path where the license file is stored.

    Methods:
        cache: Returns the current license data stored in memory.
        license_path: Returns the file path for the license file.
        save_licensefile(): Encrypts and saves the license data to a file.
        load_licensefile(): Loads and decrypts the license data from a file.
        set_path(product_code, custom_path): Sets the file path for the license file based on the operating system.
        delete_licensefile(): Deletes the license file and clears the license data in memory.
        create_activation_guard(id): Creates OfflineActivationGuard object.
        create_request_file(req_data,offline_path): Creates .req file for license activation or deactivation.
        save_guard_file(guard): Saves guard file locally.
        remove_offline_activation_data(activation_file): Removes guard and activation file.
        load_guard_file(): Loads locally saved guard file.
        check_guard(offline_data): Checks the validity of the .lic file.
        load_offline_response(file_path): Loads offline response file.
    """

    def __init__(self, conf: Configuration):
        self._conf = conf
        super().__init__(self._conf._file_key, self._conf._file_iv)

        self._cache = LicenseData(
            product=self._conf._product,
            hardwareID=self._conf._hardware_id_provider.get_id(self),
            grace_period_conf=self._conf.grace_period_conf,
        )

        self._filename = self._conf._filename
        self._license_path = self.set_path(self._conf._product, self._conf._file_path)

    @property
    def cache(self):
        return self._cache.to_json()

    @property
    def license_path(self):
        return self._license_path

    def save_licensefile(self):
        """
        Creates path for licensefile
        Saves encrypted string of licensefile JSON
        """

        json_string_encrypted = self.encrypt(self._cache.to_json())
        with self._cache._lock:
            if not os.path.exists(self.license_path):
                os.makedirs(self.license_path)

            with open(
                os.path.join(self.license_path, self._filename + ".key"), "w"
            ) as file:
                file.write(json_string_encrypted)

    def create_activation_guard(self, id: str) -> OfflineActivationGuard:
        """
        Creates OfflineActivationGuard object

        Args:
            id (str): response_id

        Returns:
            OfflineActivationGuard
        """

        offline_guard = OfflineActivationGuard()
        offline_guard.set_id(id=id)
        offline_guard.set_device_id(self._conf._hardware_id_provider.get_id(self))
        offline_guard.set_date_created()

        return offline_guard

    def create_request_file(
        self, req_data: OfflineActivation, offline_path: str
    ) -> str:
        """
        Creates .req file for license activation or deactivation

        Args:
            req_data (OfflineActivation): OfflineActivation object
            offline_path (str): sets a path for .req file

        Returns:
            str: path of a .req file
        """

        if not os.path.exists(offline_path):
            os.makedirs(offline_path)

        filename = (
            "activate_offline.req"
            if req_data._is_activation
            else "deactivate_offline.req"
        )

        with open(os.path.join(offline_path, filename), mode="w") as f:
            print(req_data._data, file=f)

        if req_data._is_activation:
            self.save_guard_file(req_data._guard)

        return os.path.join(offline_path, filename)

    def save_guard_file(self, guard: OfflineActivationGuard) -> None:
        """
        Saves guard file locally

        Args:
            guard (OfflineActivationGuard): OfflineActivationGuard object
        """

        if self._conf.is_guard_file_enabled:
            if not os.path.exists(self.license_path):
                os.makedirs(self.license_path)
            self.remove_offline_activation_data()
            with open(
                os.path.join(self.license_path, "OfflineActivation.guard"), "w"
            ) as file:
                file.write(self.encrypt(guard.to_json()))

    def remove_offline_activation_data(self, activation_file: str = None):
        """
        Removes guard and activation file.

        Args:
            activation_file (str, optional): Path of the activation file. Defaults to None.
        """

        file_path_guard = os.path.join(self.license_path, "OfflineActivation.guard")

        if os.path.exists(file_path_guard):
            os.remove(file_path_guard)

        if activation_file != None:
            if os.path.exists(activation_file):
                os.remove(activation_file)

    def load_guard_file(self) -> dict:
        """
        Loads locally saved guard file

        Returns:
            dict: guard file
        """

        with open(
            os.path.join(self.license_path, "OfflineActivation.guard"), "r"
        ) as file:
            json_string_encrypted = file.read()

        return json.loads(self.decrypt(json_string_encrypted))

    def check_guard(self, offline_data: OfflineActivation):
        """
        Checks the validity of the .lic file

        Args:
            offline_data (OfflineActivation): OfflineActivation object

        Raises:
            LicenseActivationException: Activation data is not valid
            LicenseActivationException: Response file ID mismatch
            LicenseActivationException: License does not belong to this device
        """

        decoded_data = offline_data.decode_offline_activation()

        if offline_data._guard._date_created > datetime.now(timezone.utc).replace(
            tzinfo=None
        ):
            raise LicenseActivationException(
                ErrorType.OFFLINE_ACTIVATION_ERROR,
                "Activation data is not valid, please restart activation process.",
            )

        elif offline_data._guard._id != decoded_data["request_id"]:
            raise LicenseActivationException(
                ErrorType.OFFLINE_ACTIVATION_ERROR,
                "Response file ID mismatch, please restart activation process.",
            )

        elif offline_data._guard._device_id != self._conf._hardware_id_provider.get_id(
            self
        ):
            raise LicenseActivationException(
                ErrorType.OFFLINE_ACTIVATION_ERROR,
                "License does not belong to this device.",
            )

    def load_offline_response(self, file_path: str) -> str:
        """
        Loads offline response file

        Args:
            file_path (str): file path

        Returns:
            str: string data
        """

        file_path = os.path.join(file_path)

        with open(file_path, "r") as file:
            data = file.read()

        return data

    def load_licensefile(self) -> dict:
        """
        Loads and decrypts licensefile

        Returns:
            dict: licensefile
        """

        try:
            with open(
                os.path.join(self.license_path, self._filename + ".key"), "r"
            ) as file:
                json_string_encrypted = file.read()

            licensefile_dict = json.loads(self.decrypt(json_string_encrypted))

            self._cache.from_json_to_attr(licensefile_dict)

            self._cache.grace_period_conf = self._conf.grace_period_conf

            return licensefile_dict

        except FileNotFoundError:
            raise LicenseDeleted(ErrorType.NO_LICENSEFILE, "Licensefile doesn't exist")

    def set_path(self, product_code: str, custom_path=None) -> str:
        """
        Set path for licensefile

        Parameters:
            product_code (str): short product code of LicenseSpring product
            custom_path(str,optional): custom path of licensefile

        Returns:
            str: Path of licensefile
        """

        if custom_path is not None:
            return custom_path

        if os.name == "nt":  # Windows
            base_path = os.path.join(
                os.environ.get("SystemDrive"), "Users", os.environ.get("USERNAME")
            )
            return os.path.join(
                base_path, "AppData", "Local", "LicenseSpring", product_code
            )

        elif os.name == "posix":  # Linux and macOS
            if "HOME" in os.environ:
                base_path = os.environ["HOME"]
                return os.path.join(
                    base_path, ".LicenseSpring", "LicenseSpring", product_code
                )

            else:  # macOS and other POSIX systems
                base_path = os.path.expanduser("~")
                return os.path.join(
                    base_path,
                    "Library",
                    "Application Support",
                    "LicenseSpring",
                    product_code,
                )

        else:
            raise Exception("Unsupported operating system")

    def delete_licensefile(self):
        """
        Permanently deletes licensefile and clears cache

        Returns: None
        """

        self.remove_offline_activation_data()

        if os.path.exists(os.path.join(self.license_path)):
            os.remove(os.path.join(self.license_path, self._filename + ".key"))

        self._cache = LicenseData(
            product=self._conf._product,
            hardwareID=self._conf._hardware_id_provider.get_id(self),
            grace_period_conf=self._conf.grace_period_conf,
        )



import logging
import threading
from datetime import datetime, timezone
from typing import Callable

from licensespring.licensefile.error import ErrorType, WatchdogException

MIN_WATCHDOG_TIMEOUT = 1  # Minimum timeout in minutes
DEFAULT_TIMEOUT_NF = 60  # Default timeout for non-floating licenses in minutes
TIMEOUT_ADJUSTMENT = 15  # Timeout adjustment for floating licenses in seconds


class LicenseWatchdog:
    """
    Monitors and ensures the validity of a software license in a background thread. This watchdog
    can automatically adjust its checking interval based on the type of license (floating or non-floating)
    and supports immediate or delayed license checking after being started. It also allows running
    the monitoring thread as either a daemon or a normal thread.

    Attributes:
        license: The license object being monitored.
        watchdog_callback: A callable function to be called when an exception occurs during license checking.
        timeout: The initial timeout value in minutes for license checking intervals.
        timeout_secs: The calculated timeout in seconds, adjusted for floating licenses if necessary.
        force_exit: A flag indicating whether the watchdog thread should exit its loop and stop.
        predicate: A condition variable predicate used to control the timing of license checks.
        mutex: A threading lock to ensure thread-safe operations.
        cv: A threading condition variable for synchronizing the license checking operations.
        watchdog_thread: The background thread that performs the license checking.

    Methods:
        __init__(self, license, callback: Callable, timeout: int):
            Initializes the LicenseWatchdog with a license object, a callback function, and a timeout value.

        is_active(self):
            Checks if the watchdog's background thread is currently active and running.

        run(self, run_immediately: bool = True, daemon: bool = False):
            Starts the license monitoring process in a background thread. Allows for immediate or delayed start
            of the license checking and sets the thread as a daemon based on the `daemon` argument.

        stop(self):
            Stops the license monitoring process and ensures the background thread is gracefully terminated.

        watchdog_func(self):
            The main function executed by the background thread that periodically checks the license's validity,
            waits for the specified interval or until a stop signal is received, and calls the provided callback function
            in case of exceptions during license checking.
    """

    def __init__(self, license, callback: Callable, timeout: int):
        self.license = license
        self.watchdog_callback = callback
        self.timeout = timeout
        self.timeout_secs = 0
        self.force_exit = False
        self.predicate = False
        self.mutex = threading.Lock()
        self.cv = threading.Condition(self.mutex)
        self.watchdog_thread = None

        need_adjust = False

        if self.timeout <= 0:
            if (
                self.license.licensefile_handler._cache.is_floating
                or self.license.licensefile_handler._cache.is_floating_cloud
            ):
                need_adjust = True
                self.timeout = self.license.licensefile_handler._cache.floating_timeout
                if self.timeout < MIN_WATCHDOG_TIMEOUT:
                    self.timeout = MIN_WATCHDOG_TIMEOUT
            else:
                self.timeout = DEFAULT_TIMEOUT_NF

        self.timeout_secs = self.timeout * 60  # Convert to seconds
        if need_adjust:
            self.timeout_secs -= (
                TIMEOUT_ADJUSTMENT  # Adjust timeout for license check call
            )

    def is_active(self):
        """
        Determines whether the watchdog's background thread is currently active.

        Returns:
            bool: True if the watchdog thread is alive and monitoring the license, False otherwise.
        """

        return self.watchdog_thread and self.watchdog_thread.is_alive()

    def run(self, run_immediately: bool = True, deamon: bool = False):
        """
        Starts the license monitoring process in a separate background thread.

        Args:
            run_immediately: If True, the license check is performed immediately after the thread starts.
                             Otherwise, the first check will occur after the specified timeout period.
            daemon: If True, the background thread is marked as a daemon thread, meaning it will not prevent the
                    program from exiting if it's the only thread left running. If False, the thread is a normal
                    thread that the program will wait for before exiting.

        This method creates and starts a new thread targeting the `watchdog_func` method. It also initializes
        control flags used by the thread for scheduling checks and handling thread termination.
        """

        if self.is_active():
            return

        try:
            self.watchdog_thread = threading.Thread(
                target=self.watchdog_func, daemon=deamon
            )
            self.watchdog_thread.start()
            logging.info("LicenseWatchdog run: thread created")

            with self.mutex:
                self.predicate = run_immediately
                self.force_exit = False

                self.cv.notify_all()

        except Exception as ex:
            logging.info(f"LicenseWatchdog run: exception: {ex}")
            raise Exception(f"Failed to create watchdog thread: {str(ex)}")

    def stop(self):
        """
        Signals the background thread to stop monitoring and exit gracefully.

        This method sets a flag indicating that the thread should stop after completing its current iteration
        of license checking. It then waits for the thread to acknowledge this request and terminate, ensuring
        that all resources are cleanly released.
        """

        if not self.watchdog_thread or not self.watchdog_thread.is_alive():
            return

        with self.mutex:
            self.predicate = True
            self.force_exit = True

            self.cv.notify_all()

        self.watchdog_thread.join()
        self.watchdog_thread = None

    def watchdog_func(self):
        """
        The main function executed by the background thread to perform periodic license checks.

        This function runs in a loop, periodically waiting for a signal to proceed with a license check,
        or for a signal to exit if stopping the watchdog has been requested. It performs the license check
        by calling the `check` method on the monitored license object and handles any exceptions by invoking
        the configured callback function.

        The loop continues until an explicit stop request is received, at which point the thread exits
        cleanly, ensuring all resources are properly released.
        """

        try:
            logging.info("Watchdog started")
            while True:
                with self.cv:
                    self.cv.wait_for(lambda: self.predicate, timeout=self.timeout_secs)
                    self.predicate = False

                if self.force_exit:
                    logging.info(f"break")
                    break

                else:
                    self.license.check()

        except Exception as ex:
            self.watchdog_callback(ex)
        finally:
            logging.info("Watchdog finished")


DEFAULT_TIMEOUT = 60 * 120  # 120 minutes
MIN_TIMEOUT = 60  # 1 minute
API_CALL_SLACK = 15  # Check feature API_CALL_SLACK seconds before it expires
TIMEOUT_SLACK = (
    10  # If feature times out in n seconds, set watchdog timeout to n - TIMEOUT_SLACK
)


class FeatureWatchdog:
    def __init__(self, license, callback, timeout=DEFAULT_TIMEOUT):
        self.license = license
        self.watchdog_callback = callback
        self.timeout = timeout
        self.force_exit = False
        self.predicate = False
        self.feature_codes_and_end_dates = {}
        self.watchdog_thread = None
        self.mutex = threading.Lock()
        self.cv = threading.Condition(self.mutex)

    def __del__(self):
        self.stop()

    def run(self, run_immediately: bool = False, deamon: bool = False):
        """
        Run feature watchdog thread

        Args:
            run_immediately (bool, optional): If True, the license check is performed immediately after the thread starts.
                             Otherwise, the first check will occur after the specified timeout period. Defaults to False.
            deamon (bool, optional): If True, the background thread is marked as a daemon thread, meaning it will not prevent the
                    program from exiting if it's the only thread left running. If False, the thread is a normal
                    thread that the program will wait for before exiting. Defaults to False.

        Raises:
            WatchdogException: Error while Feature WatchDog setup
        """
        try:
            if self.is_active():
                return
            self.watchdog_thread = threading.Thread(
                target=self.watchdog_func, daemon=deamon
            )
            self.watchdog_thread.start()
            logging.info("FeatureWatchdog run: thread created")

            with self.cv:
                self.predicate = run_immediately
                self.force_exit = False
                self.cv.notify_all()

        except Exception as ex:
            logging.error(f"FeatureWatchdog run: exception: {str(ex)}")
            raise WatchdogException(
                ErrorType.FEATURE_WATCHDOG_ERROR,
                f"Failed to create watchdog thread: {str(ex)}",
            )

    def stop(self):
        """
        Signals the background thread to stop monitoring and exit gracefully.
        """
        if not self.watchdog_thread or not self.watchdog_thread.is_alive():
            return

        with self.cv:
            self.predicate = True
            self.force_exit = True
            self.cv.notify_all()

        self.watchdog_thread.join()
        self.watchdog_thread = None

    def add_feature(self, feature: str):
        """
        Add feature inside FeatureWatchdog

        Args:
            feature (str): feature code
        """
        with self.cv:
            feature_obj = self.license.licensefile_handler._cache.get_feature_object(
                feature
            )
            if feature_obj == None:
                return

            if not (
                feature_obj.is_online_floating() or feature_obj.is_offline_floating()
            ):
                return

            if feature_obj.floating_is_expired():
                return

            if feature not in self.feature_codes_and_end_dates:
                logging.info(f"FeatureWatchdog addFeature: {feature}")
                self.feature_codes_and_end_dates[
                    feature
                ] = feature_obj.floating_end_datetime()
                self.predicate = True
                self.cv.notify_all()

    def remove_feature(self, feature: str):
        """
        Remove feature inside FeatureWatchdog

        Args:
            feature (str): feature code
        """
        with self.cv:
            feature_obj = self.license.licensefile_handler._cache.get_feature_object(
                feature
            )

            if not (
                feature_obj.is_online_floating() or feature_obj.is_offline_floating()
            ):
                return

            if feature in self.feature_codes_and_end_dates:
                logging.info(f"FeatureWatchdog removeFeature: {feature}")
                del self.feature_codes_and_end_dates[feature]

    def watchdog_func(self):
        """
        The main function executed by the background thread to perform periodic feature checks.
        """
        try:
            logging.info("Feature watchdog started")
            while True:
                timeout = self.calculate_timeout()
                with self.cv:
                    self.cv.wait(timeout)
                    if self.force_exit:
                        break
                    self.predicate = False

                for (
                    feature_code,
                    end_datetime,
                ) in self.feature_codes_and_end_dates.items():
                    if self.seconds_remaining(end_datetime) <= API_CALL_SLACK:
                        logging.info(f"feature:{feature_code}, timeout:{end_datetime}")
                        self.license.check_feature(feature_code, False)
                        self.feature_codes_and_end_dates[
                            feature_code
                        ] = self.license.licensefile_handler._cache.get_feature_object(
                            feature_code
                        ).floating_end_datetime()

        except Exception as ex:
            self.watchdog_callback(ex)
        logging.info("Feature watchdog finished")

    def calculate_timeout(self) -> int:
        """
        Calculates timeout for features check

        Returns:
            int: timeout in seconds
        """
        with self.cv:
            timeout = self.timeout
            for end_datetime in self.feature_codes_and_end_dates.values():
                timeout = min(timeout, self.seconds_remaining(end_datetime))
            if timeout <= TIMEOUT_SLACK:
                return timeout
            return timeout - TIMEOUT_SLACK

    def get_feature_codes(self) -> list:
        with self.cv:
            return list(self.feature_codes_and_end_dates.keys())

    def seconds_remaining(self, end_datetime):
        remaining = (
            end_datetime - datetime.now(timezone.utc).replace(tzinfo=None)
        ).total_seconds()
        return max(0, remaining)

    def is_active(self):
        return self.watchdog_thread and self.watchdog_thread.is_alive()



import base64

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature.pkcs1_15 import PKCS115_SigScheme


class SignatureVerifier:
    PUBLIC_KEY = """-----BEGIN PUBLIC KEY-----
    MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAwEt7ZQWSzeBYNBwyi3KW
    +XU/6I+ju90R5ZpTvSE4VWL8KSgVJ6bWKIhaKTL2hbUpIQQgS5ZKfa4SEEMdIm5k
    xe0u2F64JFAVQunx/1O/UXsD7ADVt93Q/hxA9Npa16tKepZoyhi79sxpaxyy/WTd
    sZKuTLApX7bWX6/XPhaPjNyiWeeK2Naka44B+F+PwN/ey3rZUra3pBltShwy2dmK
    IxJVmprf5ttMYBB+ouqPin3VgDw5Jq1FZwpLBiOe+ogR+sHu7QYWwLq6AgC4e3Jq
    gryEvqfJr/XkvOnAdcAIZm6tK6DdnLWCy6Onrk7t0VQK5nnGF1CKE+jcAByWopcI
    VCwfYrs1UCSY1YibKXNJoyvEhgLyCC+KAgXKf8omcg1Q18XKu5XSrvrCzXLnRnaF
    NpFDjcg5AJJab78hX7qGPC+e8PjuBYwh2vtx5mFj7/c+T59JM/vXwwvW9DsnDztD
    WEDFhzGanU71NgwrZbRoNSyTW0UjbZsyJVSIX6233Ng0y5L9mDe8p8P+u1B0sXAA
    ozEKL/yG4Qu5r4LIw4iv6JVPT1xbWlH4Vc4KaN3toaf4+G0EkFy6ncvBncWifLAR
    SoWpQ5YsygDdQNVYdmMsoQ76UTuNxo3eZ0sQJjbQVBqlkcXrAdtUCpuojPRsl/Xs
    9zVOJrzZPG0I98E5quNbRkMCAwEAAQ==
    -----END PUBLIC KEY-----"""

    def __init__(self):
        rsa_public_key = RSA.importKey(self.PUBLIC_KEY)
        self.verifier = PKCS115_SigScheme(rsa_public_key)

    def verify_signature(self, message, signature):
        hash = SHA256.new(message)
        self.verifier.verify(hash, signature)

    def verify_license_signature(
        self, hardware_id, auth_value, validity_period, signature_b64
    ):
        message = f'{hardware_id}#{auth_value}#{validity_period if validity_period else ""}'.lower().encode()
        signature = base64.b64decode(signature_b64)
        self.verify_signature(message, signature)

    def verify_license_signature_v2(self, message, signature_b64):
        signature = base64.b64decode(signature_b64)
        self.verify_signature(message, signature)



import logging

import requests

from licensespring.api import APIClient
from licensespring.licensefile import License
from licensespring.licensefile.config import Configuration
from licensespring.licensefile.customer import Customer
from licensespring.licensefile.data_storage import DataStorage
from licensespring.licensefile.offline_activation_guard import (
    OfflineActivation,
    OfflineActivationGuard,
)


class LicenseID:
    """
    Represents a mechanism for license identification and activation.

    The LicenseID class provides a structured way to handle both key-based and user-based
    license activations. It encapsulates the necessary details for each activation method,
    allowing for clear and concise license management within applications.

    Attributes:
        key (str, optional): The license key for key-based activation. Defaults to None.
        username (str, optional): The username (typically an email) for user-based activation. Defaults to None.
        password (str, optional): The password associated with the username for user-based activation. Defaults to None.

    Methods:
        from_key: Class method to create a LicenseID instance for key-based activation.
        from_user: Class method to create a LicenseID instance for user-based activation.
    """

    def __init__(self) -> None:
        self.key = None
        self.username = None
        self.password = None

    @classmethod
    def from_key(cls, key):
        """
        Creates an instance of LicenseID for a key-based license.

        Parameters:
            key (str): license key

        Returns:
            LicenseID
        """
        licenseID = cls()
        licenseID.key = key
        return licenseID

    @classmethod
    def from_user(cls, username, password):
        """
        Creates an instance of LicenseID for a user-based license.

        Parameters:
            username (str): email of an user
            password (str): password

        Returns:
            LicenseID
        """
        licenseID = cls()
        licenseID.username = username
        licenseID.password = password
        return licenseID


class LicenseManager:
    """
    LicenseManager is used for license activation

    Attributes:
        api_client (APIClient): An instance of the APIClient responsible for API communication.
        _product (str): LicenseSpring product.
        licensefile_handler (DataStorage): A data storage handler for managing
            license file operations.

    Parameters:
        conf (Configuration): A configuration object containing necessary parameters
            such as API keys, hardware ID provider, and other settings required
            for license activation and verification.
    """

    def __init__(
        self,
        conf: Configuration,
    ) -> None:
        self.api_client = APIClient(
            api_key=conf._api_key,
            shared_key=conf._shared_key,
            hardware_id_provider=conf._hardware_id_provider,
            verify_license_signature=conf._verify_license_signature,
            signature_verifier=conf._signature_verifier,
            api_domain=conf._api_domain,
            api_version=conf._api_version,
        )

        self._product = conf._product

        self.licensefile_handler = DataStorage(conf=conf)

    def load_license(self) -> License:
        """
        Loads licensefile and sets attributes for LicenseData instance
        Returns:
        License: An instance of the License class reflecting the loaded license.
        """
        self.licensefile_handler.load_licensefile()

        return License(self._product, self.api_client, self.licensefile_handler)

    def activate_license(
        self,
        license_id: LicenseID,
        bundle_code: str = None,
        hardware_id: str = None,
        unique_license_id: int = None,
        customer_account_code: str = None,
        id_token: str = None,
        code: str = None,
        app_ver: str = None,
        os_ver: str = None,
        hostname: str = None,
        ip: str = None,
        is_vm: bool = None,
        vm_info: str = None,
        mac_address: str = None,
    ) -> License:
        """
        Activates a license with the license server and updates local license data accordingly.

        This method supports a variety of parameters to accommodate different licensing scenarios,
        including key-based activation, user-based activation

        Args:
            license_id (LicenseID): An instance containing the license key or user credentials.
            bundle_code (str, optional): specify unique bundle_code of a bundle
            hardware_id (str, optional): A unique identifier for the hardware.
            unique_license_id (int, optional): A unique identifier for the license.
            customer_account_code (str, optional): An account code for the customer.
            id_token (str, optional): Token for identity verification.
            code (str, optional): An additional code for license verification.
            app_ver (str, optional): The version of the application requesting activation.
            os_ver (str, optional): The operating system version of the host.
            hostname (str, optional): The hostname of the device requesting activation.
            ip (str, optional): The IP address of the device.
            is_vm (bool, optional): Indicates whether the application is running on a virtual machine.
            vm_info (str, optional): Information about the virtual machine, if applicable.
            mac_address (str, optional): The MAC address of the device.

        Returns:
            License: An instance of the License class reflecting the activated license.
        """

        response = self.api_client.activate_license(
            product=self._product,
            bundle_code=bundle_code,
            hardware_id=hardware_id,
            license_key=license_id.key,
            license_id=unique_license_id,
            username=license_id.username,
            password=license_id.password,
            customer_account_code=customer_account_code,
            id_token=id_token,
            code=code,
            app_ver=app_ver,
            os_ver=os_ver,
            hostname=hostname,
            ip=ip,
            is_vm=is_vm,
            vm_info=vm_info,
            mac_address=mac_address,
        )

        response["username"] = license_id.username
        response["bundle_code"] = bundle_code
        self.licensefile_handler._cache.update_cache("activate_license", response)
        self.licensefile_handler.save_licensefile()

        return License(
            self._product,
            api_client=self.api_client,
            licensefile_handler=self.licensefile_handler,
        )

    def is_online(self) -> bool:
        response = requests.get(
            f"{self.api_client.api_protocol}://{self.api_client.api_domain}"
        )

        if response.status_code == 200:
            return True

        return False

    def create_offline_activation_file(
        self,
        license_id: LicenseID,
        req_path: str,
        bundle_code: str = None,
        hardware_id: str = None,
        unique_license_id: int = None,
        app_ver: str = None,
        os_ver: str = None,
        hostname: str = None,
        ip: str = None,
        is_vm: bool = None,
        vm_info: str = None,
        mac_address: str = None,
        device_variables: dict = None,
    ) -> str:
        """
        Creates .req file for activation

        Args:
            license_id (LicenseID): An instance containing the license key or user credentials.
            req_path (str): Specify place where to create .req file
            bundle_code (str, optional): specify unique bundle_code of a bundle.
            hardware_id (str, optional): A unique identifier for the hardware.
            app_ver (str, optional): The version of the application requesting activation.
            os_ver (str, optional): The operating system version of the host.
            hostname (str, optional): The hostname of the device requesting activation.
            ip (str, optional):  The IP address of the device.
            is_vm (bool, optional): Indicates whether the application is running on a virtual machine.
            vm_info (str, optional): Information about the virtual machine.
            mac_address (str, optional): The MAC address of the device.

        Returns:
            str: path of the .req file
        """

        data = self.api_client.activate_offline_dump(
            product=self._product,
            bundle_code=bundle_code,
            hardware_id=hardware_id,
            license_key=license_id.key,
            license_id=unique_license_id,
            username=license_id.username,
            password=license_id.password,
            app_ver=app_ver,
            os_ver=os_ver,
            hostname=hostname,
            ip=ip,
            is_vm=is_vm,
            vm_info=vm_info,
            mac_address=mac_address,
            variables=device_variables,
        )

        offline_data = OfflineActivation()
        offline_data.set_data(data)
        offline_data.set_is_activation(True)
        offline_data.create_guard_file(
            self.licensefile_handler.create_activation_guard(
                offline_data.decode_offline_activation()["request_id"]
            )
        )

        return self.licensefile_handler.create_request_file(
            req_data=offline_data, offline_path=req_path
        )

    def activate_license_offline(self, ls_activation_path: str) -> License:
        """
        Activates a .lic

        Args:
            ls_activation_path (str): path to activation file

        Raises:
            LicenseActivationException: Activation data is not valid
            LicenseActivationException: Response file ID mismatch
            LicenseActivationException: License does not belong to this device

        Returns:
            License: License object
        """

        offline_data = OfflineActivation()
        data = self.licensefile_handler.load_offline_response(ls_activation_path)
        offline_data.set_data(data)
        offline_data.set_is_activation(True)
        offline_data.set_use_guard(self.licensefile_handler._conf.is_guard_file_enabled)

        if len(offline_data._data) == 0:
            return

        response = offline_data.decode_offline_activation()

        logging.info("Offline file activation id:", response["request_id"])

        if self.licensefile_handler._conf.is_guard_file_enabled:
            guard_dict = self.licensefile_handler.load_guard_file()
            offline_data.create_guard_file(OfflineActivationGuard.from_json(guard_dict))
            self.licensefile_handler.check_guard(offline_data)

        self.licensefile_handler._cache.update_cache(
            "activate_license_offline", response
        )
        license = License(
            self._product,
            api_client=self.api_client,
            licensefile_handler=self.licensefile_handler,
        )
        license.licensefile_handler.save_licensefile()
        self.licensefile_handler.remove_offline_activation_data(ls_activation_path)
        logging.info("Offline activation succeeded")

        return license

    def get_trial_license(
        self, customer: Customer, license_policy: str = None
    ) -> LicenseID:
        """
        Creates LiceseID for trial licenses

        Args:
            customer (Customer): Customer object
            license_policy (str, optional): license policy code. Defaults to None.

        Returns:
            LicenseID
        """

        response = self.api_client.trial_key(
            self._product,
            hardware_id=None,
            email=customer.email,
            license_policy=license_policy,
            first_name=customer.first_name,
            last_name=customer.last_name,
            phone=customer.phone,
            address=customer.address,
            postcode=customer.postcode,
            state=customer.state,
            country=customer.country,
            city=customer.city,
            reference=customer.reference,
        )

        if "license" in response:
            return LicenseID.from_key(response["license"])

        else:
            return LicenseID.from_user(
                response["license_user"], response["initial_password"]
            )
